#include "fosfft.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define MAX_N (1<<18)
#define MAX_VARIANCE (1e-29)
#define MAX_VARIANCEF (1e-12)

int main(int argc, char **argv) {
  int retval = 0;
  int n, i, mode;
  double deviation, variance, variancef;
  double complex *original, *frequency, *restored;
  float complex *originalf, *frequencyf, *restoredf;
  fosfft_t *forward, *reverse;
  fosfftf_t *forwardf, *reversef;
  forward = fosfft_init_forward(MAX_N);
  reverse = fosfft_init_reverse(MAX_N);
  forwardf = fosfftf_init_forward(MAX_N);
  reversef = fosfftf_init_reverse(MAX_N);
  original = calloc(MAX_N, sizeof(*original));
  frequency = calloc(MAX_N, sizeof(*frequency));
  restored = calloc(MAX_N, sizeof(*restored));
  originalf = calloc(MAX_N, sizeof(*originalf));
  frequencyf = calloc(MAX_N, sizeof(*frequencyf));
  restoredf = calloc(MAX_N, sizeof(*restoredf));
  for (n=1; n<=MAX_N; n*=2) {
    for (mode=0; mode<(n==1?1:2); mode++) {
      if (mode == 0) printf("Full operation with n=%i\n", i);
      else printf("Half operation with n=%i\n", i);
      for (i=0; i<n; i++) {
        original[i] = (random() + I*random()) / (double)RAND_MAX * 2 - 1;
        originalf[i] = (random() + I*random()) / (double)RAND_MAX * 2 - 1;
      }
      if (mode == 0) {
        fosfft(forward, original, frequency, n);
        fosfftf(forwardf, originalf, frequencyf, n);
      } else {
        fosfft_half(forward, original, frequency, n);
        fosfftf_half(forwardf, originalf, frequencyf, n);
      }
      fosfft(reverse, frequency, restored, n);
      fosfftf(reversef, frequencyf, restoredf, n);
      variance = 0.0; variancef = 0.0;
      if (mode == 0) {
        for (i=0; i<n; i++) {
          deviation = original[i] - restored[i] / n;
          variance += deviation * deviation;
          deviation = originalf[i] - restoredf[i] / n;
          variancef += deviation * deviation;
        }
      } else {
        for (i=0; i<n/2; i++) {
          deviation = original[i] - restored[i] / n;
          variance += deviation * deviation;
          deviation = originalf[i] - restoredf[i] / n;
          variancef += deviation * deviation;
        }
        for (i=n/2; i<n; i++) {
          deviation = 0.0 - restored[i] / n;
          variance += deviation * deviation;
          deviation = 0.0 - restoredf[i] / n;
          variancef += deviation * deviation;
        }
      }
      variance /= n;
      variancef /= n;
      printf("Average variance with double precision: %e ", variance);
      if (variance > MAX_VARIANCE) {
        puts("FAIL");
        retval = 1;
      } else {
        puts("OK");
      }
      printf("Average variance with single precision: %e ", variancef);
      if (variancef > MAX_VARIANCEF) {
        puts("FAIL");
        retval = 1;
      } else {
        puts("OK");
      }
      fflush(stdout);
    }
  }
  return retval;
}
